"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.RedisCache = void 0;
const ioredis_1 = __importDefault(require("ioredis"));
const lodash_1 = require("lodash");
const autobind_decorator_1 = require("autobind-decorator");
const promise_timeout_1 = require("promise-timeout");
const Tracer_1 = require("../Tracer");
/**
 * Features:
 * - caching with tags for tag based invalidation
 * - control whether or not to throw on error
 *
 * defaults:
 *  - prefix: no prefix
 *  - defaultResponseTimeoutInMillis: 2000
 *  - defaultExpirationInSecs: 3600
 */
class RedisCache {
    constructor(config) {
        const { port, host, prefix, defaultExpirationInSecs, defaultResponseTimeoutInMillis } = config;
        this.redis = new ioredis_1.default({ port, host, keyPrefix: prefix });
        this.defaultResponseTimeoutInMillis = defaultResponseTimeoutInMillis ?? 2000;
        this.defaultExpirationInSecs = defaultExpirationInSecs ?? 3600;
    }
    async get(key, options) {
        const timeout = options?.responseTimeoutInMillis || this.defaultResponseTimeoutInMillis;
        const throwOnError = options?.throwOnError;
        return await Tracer_1.withSpan('REDIS GET', async (span) => {
            span.setAttribute(Tracer_1.ATTRIBUTES.DB_SYSTEM, Tracer_1.VALUES.DB_SYSTEM.REDIS);
            span.setAttribute(Tracer_1.ATTRIBUTES.DB_OPERATION, 'GET');
            span.setAttribute('db.redis.key', key);
            try {
                const res = await promise_timeout_1.timeout(this.redis.get(`data:${key}`), timeout);
                span.setAttribute('db.redis.cached', Boolean(res));
                return res ? JSON.parse(res) : null;
            }
            catch (error) {
                span.setAttribute('db.redis.cached', false);
                span.recordException(error);
                if (throwOnError)
                    throw error;
                return null;
            }
        });
    }
    async set(key, data, param3, param4) {
        let tags = [];
        let ttlInSecs = this.defaultExpirationInSecs;
        let timeout = this.defaultResponseTimeoutInMillis;
        let throwOnError = false;
        if (Array.isArray(param3)) {
            tags = param3;
            if (typeof param4 === 'object' && param4 !== null) {
                ttlInSecs = lodash_1.get(param4, 'ttlInSecs') || ttlInSecs;
                timeout = lodash_1.get(param4, 'responseTimeoutInMillis') || timeout;
                throwOnError = lodash_1.get(param4, 'throwOnError') || false;
            }
        }
        else if (typeof param3 === 'object' && param3 !== null) {
            ttlInSecs = lodash_1.get(param3, 'ttlInSecs') || ttlInSecs;
            timeout = lodash_1.get(param3, 'responseTimeoutInMillis') || timeout;
            throwOnError = lodash_1.get(param3, 'throwOnError') || false;
        }
        await Tracer_1.withSpan('REDIS SET', async (span) => {
            span.setAttribute(Tracer_1.ATTRIBUTES.DB_SYSTEM, Tracer_1.VALUES.DB_SYSTEM.REDIS);
            span.setAttribute(Tracer_1.ATTRIBUTES.DB_OPERATION, 'SET');
            span.setAttribute('db.redis.key', key);
            if (tags.length > 0)
                span.setAttribute('db.redis.tags', tags.join(','));
            try {
                await promise_timeout_1.timeout((async () => {
                    const multi = await this.redis.multi();
                    tags.forEach((tag) => multi.sadd(`tags:${tag}`, key));
                    multi.set(`data:${key}`, JSON.stringify(data), 'ex', ttlInSecs);
                    await multi.exec();
                })(), timeout);
            }
            catch (error) {
                span.recordException(error);
                if (throwOnError)
                    throw error;
            }
        });
    }
    async delete(key, options) {
        const timeout = options?.responseTimeoutInMillis || this.defaultResponseTimeoutInMillis;
        const throwOnError = options?.throwOnError;
        return await Tracer_1.withSpan('REDIS DEL', async (span) => {
            span.setAttribute(Tracer_1.ATTRIBUTES.DB_SYSTEM, Tracer_1.VALUES.DB_SYSTEM.REDIS);
            span.setAttribute(Tracer_1.ATTRIBUTES.DB_OPERATION, 'DEL');
            span.setAttribute('db.redis.key', key);
            try {
                await promise_timeout_1.timeout(this.redis.del(`data:${key}`), timeout);
            }
            catch (error) {
                span.recordException(error);
                if (throwOnError)
                    throw error;
            }
        });
    }
    async invalidateTags(tags, options) {
        if (tags.length === 0)
            return;
        const timeout = options?.responseTimeoutInMillis || 15000;
        const throwOnError = options?.throwOnError;
        return await Tracer_1.withSpan('REDIS INVALIDATE TAGS', async (span) => {
            span.setAttribute(Tracer_1.ATTRIBUTES.DB_SYSTEM, Tracer_1.VALUES.DB_SYSTEM.REDIS);
            span.setAttribute(Tracer_1.ATTRIBUTES.DB_OPERATION, 'INVALIDATE TAGS');
            span.setAttribute('db.redis.tags', tags.join(','));
            try {
                await promise_timeout_1.timeout((async () => {
                    const keys = (await Promise.all(tags.map((tag) => this.redis.smembers(`tags:${tag}`)))).flatMap((e) => e);
                    const pipeline = await this.redis.pipeline();
                    keys.forEach((key) => pipeline.del(`data:${key}`));
                    tags.forEach((tag) => pipeline.del(`tags:${tag}`));
                    await pipeline.exec();
                })(), timeout);
            }
            catch (error) {
                span.recordException(error);
                if (throwOnError)
                    throw error;
            }
        });
    }
    async flushAll() {
        await this.redis.flushall();
    }
}
__decorate([
    autobind_decorator_1.boundMethod
], RedisCache.prototype, "get", null);
__decorate([
    autobind_decorator_1.boundMethod
], RedisCache.prototype, "set", null);
__decorate([
    autobind_decorator_1.boundMethod
], RedisCache.prototype, "delete", null);
__decorate([
    autobind_decorator_1.boundMethod
], RedisCache.prototype, "invalidateTags", null);
__decorate([
    autobind_decorator_1.boundMethod
], RedisCache.prototype, "flushAll", null);
exports.RedisCache = RedisCache;
